#  Copyright © 2021 CloudBlue. All rights reserved.

import json
import os
from datetime import datetime

import pytz

from openpyxl import load_workbook
from openpyxl.styles import Alignment, Font, PatternFill
from openpyxl.styles.colors import Color, WHITE
from openpyxl.utils.exceptions import InvalidFileException

from zipfile import BadZipfile

from connect.reports.renderers.base import BaseRenderer
from connect.reports.renderers.registry import register


@register('xlsx')
class XLSXRenderer(BaseRenderer):
    def render(self, data, output_file, start_time=None):
        self.start_time = start_time or datetime.now(tz=pytz.utc)
        return self.generate_report(data, output_file)

    def generate_report(self, data, output_file):
        start_col_idx = self.args.get('start_col', 1)
        row_idx = self.args.get('start_row', 2)
        wb = load_workbook(
            os.path.join(
                self.root_dir,
                self.template,
            ),
        )
        ws = wb['Data']
        for row in data:
            for col_idx, cell_value in enumerate(row, start=start_col_idx):
                ws.cell(row_idx, col_idx, value=cell_value)
            row_idx += 1

        self._add_info_sheet(wb.create_sheet('Info'), self.start_time)

        output_file = f'{output_file}.xlsx'
        wb.save(output_file)
        return output_file

    def _add_info_sheet(self, ws, start_time):
        ws.column_dimensions['A'].width = 50
        ws.column_dimensions['B'].width = 180
        ws['A1'].value = 'Report Execution Information'
        ws.merge_cells('A1:B1')
        ws['A1'].fill = PatternFill('solid', start_color=Color('1565C0'))
        ws['A1'].font = Font(sz=24, color=WHITE)
        ws['A1'].alignment = Alignment(horizontal='center', vertical='center')
        for i in range(2, 10):
            ws[f'A{i}'].alignment = Alignment(
                horizontal='left',
                vertical='top',
            )
            ws[f'B{i}'].alignment = Alignment(
                horizontal='left',
                vertical='top',
            )
        ws['A2'].value = 'Report Start time'
        ws['B2'].value = start_time.isoformat()
        ws['A3'].value = 'Report Finish time'
        ws['B3'].value = datetime.now(tz=pytz.utc).isoformat()
        ws['A4'].value = 'Account ID'
        ws['B4'].value = self.account.id
        ws['A5'].value = 'Account Name'
        ws['B5'].value = self.account.name
        ws['A6'].value = 'Report ID'
        ws['B6'].value = self.report.id
        ws['A7'].value = 'Report Name'
        ws['B7'].value = self.report.name
        ws['A8'].value = 'Runtime environment'
        ws['B8'].value = self.environment
        ws['A9'].value = 'Report execution parameters'
        ws['B9'].value = json.dumps(self.report.values, indent=4, sort_keys=True)
        ws['B9'].alignment = Alignment(
            horizontal='left',
            vertical='top',
            wrap_text=True,
        )

    @classmethod
    def _validate_args(cls, args):
        errors = []
        start_row = args.get('start_row')
        start_col = args.get('start_col')
        if start_row is not None:
            if not isinstance(start_row, int):
                errors.append('`start_row` must be integer.')
            else:
                if start_row < 1:
                    errors.append('`start_row` must be greater than 0.')
        if start_col is not None:
            if not isinstance(start_col, int):
                errors.append('`start_col` must be integer.')
            else:
                if start_col < 1:
                    errors.append('`start_col` must be greater than 0.')
        return errors

    @classmethod
    def validate(cls, definition):
        errors = []
        if definition.template is None:
            errors.append('`template` is required for xlsx renderer.')
            return errors
        elif not os.path.isfile(
            os.path.join(definition.root_path, definition.template),
        ):
            errors.append(f'template `{definition.template}` not found.')
            return errors

        try:
            template_file = os.path.join(definition.root_path, definition.template)
            load_workbook(template_file)
        except (InvalidFileException, BadZipfile):
            errors.append(f'template `{definition.template}` not valid or empty.')

        if definition.args is not None:
            errors.extend(cls._validate_args(definition.args))
        return errors



#  Copyright © 2021 CloudBlue. All rights reserved.

import os
from functools import partial

from weasyprint import HTML, default_url_fetcher

from connect.reports.renderers.j2 import Jinja2Renderer
from connect.reports.renderers.registry import register


def local_fetcher(url, root_dir=None, template_dir=None):
    base_path = os.path.join(
        os.path.abspath(root_dir),
        template_dir,
    )
    if url.startswith('file://') and not url.startswith(f'file://{base_path}'):
        cwd = os.getcwd()
        rel_path = os.path.relpath(url[7:], cwd)
        new_path = os.path.join(
            os.path.abspath(root_dir),
            template_dir,
            rel_path,
        )
        new_url = f'file://{new_path}'
        return default_url_fetcher(new_url)
    return default_url_fetcher(url)


@register('pdf')
class PDFRenderer(Jinja2Renderer):
    def generate_report(self, data, output_file):
        tokens = output_file.split('.')
        if tokens[-1] != 'pdf':
            output_file = f'{tokens[0]}.pdf'

        rendered_file = super().generate_report(data, output_file)
        fetcher = partial(
            local_fetcher,
            root_dir=self.root_dir,
            template_dir=os.path.dirname(self.template),
        )
        html = HTML(filename=rendered_file, url_fetcher=fetcher)
        html.write_pdf(output_file)
        return output_file

    @classmethod
    def validate(cls, definition):
        errors = super(PDFRenderer, cls).validate(definition)
        if definition.args is not None:
            css_file = definition.args.get('css_file')
            if css_file and not os.path.isfile(
                os.path.join(definition.root_path, css_file),
            ):
                errors.append(
                    f'css_file `{css_file}` not found.',
                )
        return errors



#  Copyright © 2021 CloudBlue. All rights reserved.

import os
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List


@dataclass
class Account:
    id: str
    name: str


@dataclass
class Report:
    id: str
    name: str
    description: str
    values: List[Dict[str, Any]]


@dataclass
class RendererDefinition:
    root_path: str
    id: str
    type: str
    description: str
    template: str = field(default=None)
    args: Dict[str, Any] = field(default=None)


@dataclass
class ParameterDefinition:
    id: str
    type: str
    name: str
    description: str
    required: bool = False


@dataclass
class ChoicesParameterDefinition(ParameterDefinition):
    choices: List[Dict[str, str]] = field(default_factory=list)


@dataclass
class ReportDefinition:
    root_path: str
    name: str
    readme_file: str
    entrypoint: str
    audience: List[str]
    report_spec: str
    default_renderer: str
    renderers: List[RendererDefinition]
    parameters: List[ParameterDefinition] = field(default_factory=list)

    @property
    def local_id(self):
        tokens = self.entrypoint.split('.')
        return tokens[1] if len(tokens) > 2 else None

    @property
    def description(self):
        path = os.path.join(self.root_path, self.readme_file)
        with open(path, 'r') as fp:
            return fp.read()

    def get_parameters(self):
        return [
            asdict(p)
            for p in self.parameters
        ]


@dataclass
class RepositoryDefinition:
    root_path: str
    readme_file: str
    name: str
    version: str
    language: str = 'python'
    reports: List[ReportDefinition] = field(default_factory=list)

    @property
    def description(self):
        path = os.path.join(self.root_path, self.readme_file)
        with open(path, 'r') as fp:
            return fp.read()



#  Copyright © 2021 CloudBlue. All rights reserved.

from connect.reports.datamodels import (
    ChoicesParameterDefinition,
    ParameterDefinition,
    RendererDefinition,
    ReportDefinition,
    RepositoryDefinition,
)


def parse(root_path, data):
    reports = data.pop('reports')
    reports_definitions = []
    for report in reports:
        parameters_definitions = []
        for param in report.pop('parameters'):
            cls = ChoicesParameterDefinition if 'choices' in param else ParameterDefinition
            parameters_definitions.append(cls(**param))

        if report['report_spec'] == '1':
            default_renderer = 'default_xlsx_renderer'
            template = report.pop('template')
            start_row = report.pop('start_row')
            start_col = report.pop('start_col')
            renderers_definitions = [
                RendererDefinition(
                    root_path=root_path,
                    id=default_renderer,
                    type='xlsx',
                    description='Render report to Excel.',
                    template=template,
                    args={
                        'start_row': start_row,
                        'start_col': start_col,
                    }),
            ]
        if report['report_spec'] == '2':
            default_renderer = report.pop('default_renderer')
            renderers_definitions = [
                RendererDefinition(root_path=root_path, **renderer)
                for renderer in report.pop('renderers')
            ]

        reports_definitions.append(
            ReportDefinition(
                root_path=root_path,
                default_renderer=default_renderer,
                parameters=parameters_definitions,
                renderers=renderers_definitions,
                **report,
            ),
        )

    return RepositoryDefinition(
        root_path=root_path,
        reports=reports_definitions,
        **data,
    )



#  Copyright © 2021 CloudBlue. All rights reserved.

import os

from jinja2 import (
    Environment,
    FileSystemLoader,
    select_autoescape,
)

from connect.reports.renderers.base import BaseRenderer
from connect.reports.renderers.registry import register


@register('jinja2')
class Jinja2Renderer(BaseRenderer):
    def generate_report(self, data, output_file):
        path, name = self.template.rsplit('/', 1)
        loader = FileSystemLoader(os.path.join(self.root_dir, path))
        env = Environment(
            loader=loader,
            autoescape=select_autoescape(['html', 'xml']),
        )
        template = env.get_template(name)
        _, ext, _ = name.rsplit('.', 2)

        report_file = f'{output_file}.{ext}'
        template.stream(self.get_context(data)).dump(open(report_file, 'w'))
        return report_file

    @classmethod
    def validate(cls, definition):
        errors = []
        if definition.template is None:
            errors.append('`template` is required for jinja2 renderer.')
        else:
            if not os.path.isfile(
                os.path.join(definition.root_path, definition.template),
            ):
                errors.append(f'template `{definition.template}` not found.')

            tokens = definition.template.rsplit('.', 2)
            if len(tokens) < 3 or tokens[-1] != 'j2':
                errors.append(
                    f'invalid template name: `{definition.template}` '
                    '(must be in the form <name>.<ext>.j2).',
                )

        return errors



#  Copyright © 2021 CloudBlue. All rights reserved.

from connect.reports.renderers.base import BaseRenderer

_RENDERERS = {}


class RendererAlreadyRegisteredError(Exception):
    pass


class RendererNotFoundError(Exception):
    pass


def register(name):
    if name in _RENDERERS:
        raise RendererAlreadyRegisteredError(f'The renderer {name} is already registered.')

    def _wrapper(cls):
        if not issubclass(cls, BaseRenderer):
            raise ValueError('The provided class must be a subclass of BaseRenderer.')

        _RENDERERS[name] = cls

        return cls
    return _wrapper


def get_renderer_class(name):
    if name not in _RENDERERS:
        raise RendererNotFoundError(f'The renderer {name} does not exist.')
    return _RENDERERS[name]


def get_renderer(name, environment, project_dir, account, report, template=None, args=None):
    cls = get_renderer_class(name)
    return cls(environment, project_dir, account, report, template, args)


def get_renderers():
    return _RENDERERS.keys()



#  Copyright © 2021 CloudBlue. All rights reserved.

import json
import os
from collections import Counter

import jsonschema

from connect.reports.renderers import get_renderer_class, get_renderers


CHOICES_PARAM_TYPE = [
    'checkbox',
    'choice',
]

ALL_PARAM_TYPES = [
    'date_range',
    'date',
    'single_line',
    'object',
    'product',
    'marketplace',
    'hub',
    *CHOICES_PARAM_TYPE,
]

JSON_REPORTS_SCHEMA = os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    'schemas/reports_schema.json',
)


def _get_duplicates(data):
    return [
        item for item, count in Counter(data).items() if count > 1
    ]


def validate_with_schema(json_data):
    with open(JSON_REPORTS_SCHEMA, 'r') as fp:
        json_schema = json.load(fp)
    try:
        jsonschema.validate(instance=json_data, schema=json_schema)
    except jsonschema.exceptions.ValidationError as errors:
        return str(errors)


def _validate_parameters(parameters):
    errors = []
    params_ids = []
    for param in parameters:
        params_ids.append(param.id)

    if param.type not in ALL_PARAM_TYPES:
        errors.append(f'Invalid type for parameter {param.id}: {param.type}.')

    if param.type in CHOICES_PARAM_TYPE and not param.choices:
        errors.append(f'Missing choices for parameter {param.id}.')

    diff = set(_get_duplicates(params_ids))
    if diff:
        errors.append(
            f'The following parameter ids are duplicated: {",".join(diff)}',
        )
    return errors


def _validate_renderer(report, renderer):
    errors = []
    if renderer.type not in get_renderers():
        errors.append(
            f'renderer `{renderer.id}` of type `{renderer.type}` is not known.',
        )
        return errors

    renderer_cls = get_renderer_class(renderer.type)
    errors.extend(renderer_cls.validate(renderer))
    return errors


def _validate_report(report):
    errors = []
    if not os.path.isfile(
        os.path.join(report.root_path, report.readme_file),
    ):
        errors.append(
            f'report `{report.local_id}` property `readme_file` cannot be resolved to a file.',
        )

    tokens = report.entrypoint.split('.')
    if len(tokens) < 2:
        errors.append(
            f'entrypoint `{report.entrypoint}` does not follow the package structure.',
        )
        return errors

    report_root = os.path.join(report.root_path, tokens[0], tokens[1])
    if not (os.path.isdir(report_root) or os.path.isfile(f'{report_root}.py')):
        errors.append(
            f'entrypoint `{report.entrypoint}` directory structure '
            'does not match the package definition.',
        )

    renderers_ids = []
    for renderer in report.renderers:
        renderers_ids.append(renderer.id)
        errors.extend(_validate_renderer(report, renderer))

    diff = set(_get_duplicates(renderers_ids))
    if diff:
        errors.append(f'report {report.local_id} has duplicated renderer ids: {",".join(diff)}')

    if report.default_renderer not in renderers_ids:
        errors.append(
            f'default renderer for report {report.local_id} '
            f'does not exist: {report.default_renderer}',
        )
    return errors


def validate(repo):
    errors = []
    if not os.path.isfile(
        os.path.join(repo.root_path, repo.readme_file),
    ):
        errors.append(
            'repository property `readme_file` cannot be resolved to a file.',
        )

    reports_local_ids = []
    for report in repo.reports:
        reports_local_ids.append(report.local_id)
        errors.extend(_validate_report(report))

    diff = set(_get_duplicates(reports_local_ids))
    if diff:
        errors.append(
            f'Multiple reports within single module found: {",".join(diff)}',
        )

    return errors



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['connect', 'connect.reports', 'connect.reports.renderers']

package_data = \
{'': ['*'], 'connect.reports': ['schemas/*']}

install_requires = \
['Jinja2>=2.11.3,<3.0.0',
 'WeasyPrint>=52.2,<53.0',
 'jsonschema>=3.2.0,<4.0.0',
 'lxml>=4.6.2,<5.0.0',
 'openpyxl>=2.5.14',
 'pytz>=2021.1,<2022.0']

setup_kwargs = {
    'name': 'connect-reports-core',
    'version': '23.0.1',
    'description': 'Connect Reports Core',
    'long_description': '# Connect Reports Core\n\n![pyversions](https://img.shields.io/pypi/pyversions/connect-reports-core.svg) [![PyPi Status](https://img.shields.io/pypi/v/connect-reports-core.svg)](https://pypi.org/project/connect-reports-core/) [![Build Connect Reports Core](https://github.com/cloudblue/connect-reports-core/actions/workflows/build.yml/badge.svg)](https://github.com/cloudblue/connect-reports-core/actions/workflows/build.yml) [![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=connect-reports-core&metric=alert_status)](https://sonarcloud.io/dashboard?id=connect-reports-core) [![Coverage](https://sonarcloud.io/api/project_badges/measure?project=connect-reports-core&metric=coverage)](https://sonarcloud.io/dashboard?id=connect-reports-core) [![Maintainability Rating](https://sonarcloud.io/api/project_badges/measure?project=connect-reports-core&metric=sqale_rating)](https://sonarcloud.io/dashboard?id=connect-reports-core)\n\n## Introduction\n\n`Connect Reports Core` is the kernel package for handling reports on CloudBlue Connect Ecosystem. \nThis library is reponsible for validation of reports definition, choosing of renderer for parsing process and writing results of reports execution.\n\n\n## Install\n\n`Connect Reports Core` requires python 3.8 or later and has the following dependencies:\n\n* openpyxl>=2.5.14\n* WeasyPrint>=52.2\n* Jinja2>=2.11.3\n* jsonschema<=3.2.0\n* pytz>=2021.1\n* lxml>=4.6.2\n\n`Connect Reports Core` can be installed from [pypi.org](https://pypi.org/project/connect-reports-core/) using pip:\n\n```\n$ pip install connect-reports-core\n```\n\n## Testing\n\nOn MacOs:\n* Install system dependencies\n```commandline\nbrew install py3cairo pango\n```\n* Create virtualenv\n* Install project dependencies\n```commandline\npip install poetry\npoetry update\n```\n* Run tests\n```commandline\npoetry run pytest\n```\n\n\n## License\n\n``Connect Reports Core`` is released under the [Apache License Version 2.0](https://www.apache.org/licenses/LICENSE-2.0).\n',
    'author': 'CloudBlue LLC',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://connect.cloudblue.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



#  Copyright © 2021 CloudBlue. All rights reserved.

from connect.reports.renderers.base import BaseRenderer
from connect.reports.renderers.registry import register

import json


@register('json')
class JSONRenderer(BaseRenderer):
    def generate_report(self, data, output_file):
        tokens = output_file.split('.')
        if tokens[-1] != 'json':
            output_file = f'{tokens[0]}.json'
        json.dump(data, open(output_file, 'w'))
        return output_file
